#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 23 17:53:08 2022

@author: benjamin
"""

import numpy as np

def rmse(generated, observed, relative=False):    
    if relative:
        return np.mean(np.abs((generated-observed)/observed))
    else:
        return np.mean(np.abs((generated-observed)))
    
def kinetic(articulators):
    return np.mean(np.linalg.norm(np.diff(articulators, axis=0), axis=1)**2)

def potential(articulators, reference=0):
    return np.mean(np.linalg.norm(articulators - reference, axis=1)**2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:59:33 2022

@author: benjamin
"""

import numpy as np
from scipy.optimize import minimize, dual_annealing
from tqdm import tqdm

class Solution:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=pointless-string-statement

    error = None
    error_formants = None
    sparsity = None
    articulatory_params = None
    estimated_formants = None

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)

def cost_function(generated, observed, x=None, lbd=0):
    
    return (np.sqrt(np.mean(((generated-observed)/observed)**2)).reshape(-1).squeeze() + 
        lbd * l12norm(x))

def initial_solution(israndm=True, nb_param=7, nb_frame=1):
    if israndm:
        x0 = np.random.randn(nb_param * nb_frame)
        while (abs(x0) > 3).any():
            x0 = np.random.randn(nb_param * nb_frame)
    else:
        x0 = np.zeros(nb_param * nb_frame)
    return x0

def l12norm(x):
    if x is not None:
        nb_row, nb_col = x.shape
        if nb_col > 1:
            s = 0
            for r in range(nb_row):
                s += np.linalg.norm(x[r, :].squeeze(), 2)
        else:
            s = np.linalg.norm(x, 1)
        return s
    else:
        return 0

def mat2vec(x, nb_param, nb_frame):
    return x.reshape((nb_param*nb_frame, 1), order="F")

def multiple_optimization(args, maxiter=200, method='nelder-mead', nb_optim=10,
                          israndm=True, disable=False, nb_param=7, 
                          verbose=False):
    solutions = []
    nb_frame = args[0].shape[1]
    for n in tqdm(range(nb_optim), disable=disable):
        solution = optimization(args, x0=initial_solution(israndm, 
                                                          nb_frame=nb_frame), 
                                maxiter=maxiter, method=method)
        solutions.append(solution)
    return solutions[np.argmin([s.error for s in solutions])]

def objective(x, *args):
    """ Objective function """
    
    nb_frame = args[0].shape[1]
    gen_formants = np.zeros_like(args[0])
    nb_param = int(len(x) / nb_frame)
    x = vec2mat(x, nb_param, nb_frame)
    x0 = np.vstack((x, np.zeros((1, nb_frame))))
    gen_formants = model_to_formants(x0).squeeze()
    verbose = False
    if len(args) == 1:
        obs_formants = args[0]
        lbd = 0
    elif len(args) == 2:
        (obs_formants, lbd) = args
    elif len(args) == 3:
        (obs_formants, lbd, verbose) = args
    E = cost_function(gen_formants, obs_formants, x, lbd)
    if verbose:
        print(E)
    return E

def optimization(args, x0=np.zeros(8), maxiter=200, method='nelder-mead'):    
    options={'maxiter': maxiter * len(x0)}
    bounds = ((((-3, 3),)*len(x0)))
    try:
        if method != "annealing":
            tmp_sol = minimize(objective, x0, args=args, 
                               method=method, options=options, bounds=bounds)
        else:
            tmp_sol = dual_annealing(objective, bounds, x0=x0, 
                                  args=args, maxiter=maxiter)        
       
        x_sol = tmp_sol["x"]
        nb_frame = args[0].shape[1]
        gen_formants = np.zeros_like(args[0])
        nb_param = int(len(x_sol) / nb_frame)
        x = vec2mat(x_sol, nb_param, nb_frame)
        x0 = np.vstack((x, np.zeros((1, nb_frame))))
        gen_formants = model_to_formants(x0).squeeze()
        if len(args) == 1:
            obs_formants = args[0]
            lbd = 0
        else:
            (obs_formants, lbd) = args
        return Solution("error", cost_function(gen_formants, obs_formants, x, lbd),
                        "articulatory_params", vec2mat(x_sol, nb_param, nb_frame),
                        "estimated_formants",  gen_formants,
                        "error_formants", cost_function(gen_formants, obs_formants, 
                                                        x, 0),
                        "sparsity", l12norm(x))
    except:
        return Solution("error", np.inf)
    
def vec2mat(x, nb_param, nb_frame):
    return x.reshape((nb_param, nb_frame), order="F")
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun  3 12:42:13 2022

@author: benjamin
"""

import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import keras
from keras import layers, models, regularizers
import h5py
import numpy as np
from tqdm import tqdm
from maedeep._config import get_maedeep_path
from maedeep._areafunction import AreaFunction
from maedeep._contour import Contour
import warnings

def build_model(input_layer_size, hidden_layer_size,
                encoding_function="relu",
                decoding_function="relu",
                sparsity=0,
                loss="mean_squared_error"):

    input_art = keras.Input(shape=(input_layer_size,))
    if sparsity > 0:
        encoded = layers.Dense(hidden_layer_size, 
                               activation=encoding_function,
                               activity_regularizer=regularizers.l1(sparsity))(input_art)
    else:
        encoded = layers.Dense(hidden_layer_size, 
                               activation=encoding_function)(input_art)
    # decoded = layers.Dense(input_layer_size, 
    #                        activation=decoding_function)(encoded)
    # autoencoder = keras.Model(input_art, decoded)

    encoder = keras.Model(input_art, encoded)
    # encoded_input = keras.Input(shape=(input_layer_size,))
    # decoder_layer = autoencoder.layers[-1]
    # decoder = keras.Model(encoded_input, decoder_layer(encoded_input))
    encoder.compile(optimizer='adam', loss=loss)
    # for m in [autoencoder, encoder, decoder]:
    #     m.compile(optimizer='adam', loss='mean_squared_error')

    return encoder#, decoder, autoencoder

def build_dnn_model(nb_feat_input, nb_feat_output, 
                    hidden_layer_size=(40, 80, 40), 
                    activation_function="sigmoid",
                    loss_function="mean_squared_error"):
    
    encoding_function = activation_function
    
    input_art = keras.Input(shape=(nb_feat_input,))
    nb_layers = len(hidden_layer_size)
    
    hidden_layers = [layers.Dense(hidden_layer_size[0], 
                                activation=activation_function)(input_art)]
    for n in range(1, nb_layers):
        hidden_layers.append(layers.Dense(hidden_layer_size[n], 
                                activation=encoding_function)(hidden_layers[n-1]))
    # hidden_2 = layers.Dense(hidden_layer_size[2], 
    #                         activation=encoding_function)(hidden_1)
    decoded = layers.Dense(nb_feat_output, 
                            activation=encoding_function)(hidden_layers[-1])
    encoder = keras.Model(input_art, decoded)
    
    encoder.compile(optimizer='adam', loss=loss_function)
    return encoder


def check_input_area(area):
    if hasattr(area, "area") and hasattr(area, "length"):
        af, lf = [getattr(area, key) for key in ["area", "length"]]
        return np.vstack((af, lf))
    else:
        return area

def concat_data(x, y):
    if x is None:
        return y
    else:
        return np.hstack((x, y))
    
def dnn_mapping(x, input_space, output_space, model_directory=None, 
                model=None, norms=None):
    
    if model is None:
        model, model_file = load_mapping_model(model_directory, input_space,
                                   output_space)
    nb_feat = model.get_config()["layers"][0]["config"]["batch_input_shape"][1]
    
    if norms is None:
        with h5py.File(model_file, "r") as hf:
            norms_input = hf["normalization_factors_input"][()]
            norms_output = hf["normalization_factors_output"][()]
    else:
        norms_input, norms_output = norms
    x = normalization_model(reshape_input(x, nb_feat), norms_input, 0)
    
    return normalization_model(model.predict(x, verbose=0), norms_output, 1).T
    
def load_data(file):
    keys = ["train_input", "output_train",
            "test_input", "test_output"]
    with h5py.File(file, "r") as hf:
        data = [hf[key][()] for key in keys]
    return data

def load_model(file):
    return models.load_model(file)

def load_mapping_model(model_directory, input_space, output_space):
    if model_directory is None:
        model_directory = os.path.join(get_maedeep_path(), "dnn", "models")
    file = os.path.join(model_directory, 
                        input_space + "_to_" + output_space + ".h5")
    return load_model(file), file

def make_data_notrain(training_path, input_vars="articulators", 
              output_vars="tract", disable=False):

    list_files = [os.path.join(training_path, x) for x in os.listdir(training_path)
                  if ".h5" in x]
    x_train = None
    y_train = None
    for l in tqdm(list_files, disable=disable):
        with h5py.File(l, "r") as hf:
            x = hf[input_vars][()]
            y = hf[output_vars][()]
            rejected = hf["rejection"][()]

        if len(rejected) > 0:
            x, y = remove_rejected(x, y, rejected)
        x_train = concat_data(x_train, x)
        y_train = concat_data(y_train, y)
        
    return x_train, y_train

def make_data(training_path, input_vars="articulators", 
              output_vars="tract", prop=10, save=None, disable=False):

    list_files = [os.path.join(training_path, x) for x in os.listdir(training_path)
                  if ".h5" in x]
    x_train = None
    y_train = None
    for l in tqdm(list_files, disable=disable):
        with h5py.File(l, "r") as hf:
            x = hf[input_vars][()]
            y = hf[output_vars][()]
            rejected = hf["rejection"][()]

        if len(rejected) > 0:
            x, y = remove_rejected(x, y, rejected)
        x_train = concat_data(x_train, x)
        y_train = concat_data(y_train, y)
        
    x_train, y_train = make_data_notrain(training_path, input_vars=input_vars, 
                  output_vars=output_vars, disable=disable)
        
    x_train, y_train, x_test, y_test = make_test(x_train, y_train, prop=prop)
    if save is not None:
        save_data([x_train, y_train, x_test, y_test], save)
    return x_train, y_train, x_test, y_test

def make_test(x_train, y_train, prop_test=5, prop_valid=5):
    
    x_train, y_train, x_test, y_test = pick_random_sample(x_train, 
                                                          y_train, 
                                                          prop_test)  
    x_train, y_train, x_valid, y_valid = pick_random_sample(x_train, 
                                                          y_train, 
                                                          prop_valid) 
    
    return x_train.T, y_train.T, x_test.T, y_test.T, x_valid.T, y_valid.T
            
def mapping(x, input_space, output_space, model_dir="./"):
    
    model_file = os.path.join(model_dir, input_space + "_to_" + 
                              output_space + ".h5")
    model = load_model(model_file)
    with h5py.File(model_file, "r") as hf:
        norms_input = hf["normalization_factors_input"][()]
        norms_output = hf["normalization_factors_output"][()]

    x = normalization_model(x, norms_input, 0)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        return normalization_model(model.predict(x, verbose=0), norms_output, 1)    

def normalization_model(x, norms, direction):
    if isinstance(norms, np.int64):
        add_norm = 0
        mult_norm = norms
    elif len(norms) == 1:
        add_norm = 0
        mult_norm = norms[0]
    else:            
        add_norm = norms[1]
        mult_norm = norms[0]
    if direction == 0:
        return (x - add_norm) / mult_norm
    else:
        return x * mult_norm + add_norm

def normalization(x, axis=0):
    return (x - np.mean(x, axis=axis, keepdims=True))/ np.std(x, axis=axis, keepdims=True)

def normalization_max(x, axis=0):
    
    minx = np.min(x, axis=axis, keepdims=True)
    mx = np.max(x - minx, axis=axis, keepdims=True)
    return (x - minx) / mx, mx, minx

def normalization_tmp(x, axis=0):
    mx = np.mean(x, axis=axis, keepdims=True)
    stdx = np.std(x, axis=axis, keepdims=True)
    return (x - mx)/ stdx, mx, stdx

def normalized_data(x_train, y_train, x_test, y_test):
    return (normalization(x_train, axis=0), normalization(y_train, axis=0),
            normalization(x_test, axis=0), normalization(y_test, axis=0))

def output_area(areas, output_type):
    
    if output_type == "raw":
        return areas
    else:
        af, lf = [areas[n*40:(n+1)*40,:] for n in range(2)]
        if output_type == "separate":
            return af, lf
        elif output_type == "area_function":
            return AreaFunction("area", af, "length", lf)
        else:
            raise ValueError(output_type + """ is not a valid output type. 
                             Choose between 'raw', 'separate', 
                             and 'area_function'""")
                             
def output_contour(predicted_contours, output_type):
    
    if output_type == "raw":
        return predicted_contours.T
    else:
        low_x, low_y, up_x, up_y = [predicted_contours.T[n*29:(n+1)*29,:] for n in range(4)]
        if output_type == "separate":
            return low_x, low_y, up_x, up_y
        elif output_type == "contour":
            return [Contour("contours", [low_x[:,i], low_y[:,i],
                                        up_x[:,i], up_y[:,i]]) for i in range(low_x.shape[1])]
        else:
            raise ValueError(output_type + """ is not a valid output type. 
                             Choose between 'raw', 'separate', 
                             and 'contour'""")

def pick_random_sample(x, y, prop):
    
    nb_smpl = x.shape[1]
    nb_test = int(prop/100*nb_smpl)
    idx = np.random.randint(low=0, high=nb_smpl, size=nb_test)
    x_test = x[:, np.unique(idx)]
    y_test = y[:, np.unique(idx)]    
    
    return (np.delete(x, np.unique(idx), axis=1),
            np.delete(y, np.unique(idx), axis=1),
            x_test,
            y_test)

def remove_invariants(train, test):
    
    values = None
    idx = [x for x in range(train.shape[1]) if np.std(train[:,x]) == 0]
    if len(idx) > 0:
        values = train[0, idx]
        train = np.delete(train, idx, axis=1)
        test = np.delete(test, idx, axis=1)
        
    return train, test, (idx, values)

def remove_rejected(art_params, tract_vars, rejected):
    return (np.delete(art_params, rejected, axis=1),
            np.delete(tract_vars, rejected, axis=1))

def reshape_input(x, nb_feat):
    if isinstance(x, list):
        x = np.array(x)
    if len(x.shape) == 1:
        x = x.reshape(1, -1)
    if x.shape[1] != nb_feat:
        x = x.T
        
    return x

def save_data(values, output_file):
    
    keys = ["train_input", "output_train",
            "test_input", "test_output"]
    with h5py.File(output_file, "w") as hf:
        for key, value in zip(keys, values):
            hf.create_dataset(key, data=value)
            
def save_model(model, file):
    model.save(file)





# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 09:55:08 2020

@author: benjamin
"""

import json
import numpy as np
import numpy.matlib as npmt
from .signaltools import ChainMatrix, esprit, symmetricifft
from ._set import struct, timestruct
from tqdm import tqdm


class Waveguide:
    """ Class for analyzing and make operations on vocal tract cavities """
    #### Properties
    area_function = None # area function
    formants = None # formants
    child_wvg = None # name of the child waveguide object
    parent_wvg = None # name of the parent waveguide object
    twin_wvg = None # name of the twin waveguide object
    anabranch_wvg = None # name of the anabranch waveguide object
    child_point = None # point of connection where the child waveguide is connected
    parent_point = None # point of connection of the parent waveguide
    radiation = None # Boolean. True if the waveguide radiates (e.g. oral tract).
    transfer_function = None # transfer function of the waveguide
    transfer_function_constriction = None
    chain_matrix = None
    chain_matrix_derivative = None
    input_impedance = None
    freq = None # frequency vector of the transfer function
    actualVT = None
    velopharyngeal_port = 0
    velum_area = 0
    norder = 1

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            if key == 'area_function':
                self.area_function.parent = self

    def computetransferfunction(self, param=None, meth='tmm', 
                                loc=-1, formants=False, verbosity=False,
                                position=0, df=50):
        """ Compute the transfer function and the impulse response of the vocal tract
        defined by area function Avt and lvt. The resonance frequencies are
        estimated via ESPRIT. They are the pole frequencies of the impulse
        response of the vocal tract. The constant values are stocked in
        param.
        The transfer function is computed using the chain paradigm by Sondhi and
        Schroeter (A hybrid time-frequency domain articulatory speech
        synthesizer, IEEE TASSP, 1987)"""

        if param is None:
            param = timestruct()
        if not hasattr(param, 'rho'):
            param.rho = 1.204/1000
        if not hasattr(param, 'c'):
            param.c = 343.4
        if not hasattr(param, 'mu'):
            param.mu = 1.9831e-5
        if not hasattr(param, 'a'):
            param.a = 130 * np.pi
        if not hasattr(param, 'b'):
            param.b = (30 * np.pi)**2
        if not hasattr(param, 'c1'):
            param.c1 = 4
        if not hasattr(param, 'wo2'):
            param.wo2 = (406 * np.pi)**2
        if not hasattr(param, 'c1n'):
            param.c1n = 72
        if not hasattr(param, 'heat_cond'):
            param.heat_cond = 0.0034
        if not hasattr(param, 'specific_heat'):
            param.specific_heat = 160
        if not hasattr(param, 'adiabatic'):
            param.adiabatic = 1.4
        if not hasattr(param, 'wallyield'):
            param.wallyield = True
        if not hasattr(param, 'loss'):
            param.loss = True

        param.freq[param.freq <= 1e-11] = 1e-11
        if df != 50:
            param.freq = np.arange(0, 5000+df, df, dtype='float') #Vector of frequencies
            param.freq = param.freq[param.freq <= 5000]
        
        freq = param.freq.reshape(-1)
        af = self.area_function.area # to make it from the lips to the glottis
        lf = self.area_function.length # to make it from the lips to the glottis
        if af is None or lf is None:
            print('Warning: the waveguide instance has no area function!')
            return None
        w = 2 * np.pi * freq # angular frequency
        lw = len(w)

        if meth.lower() == 'cmp':
            param.alp = np.sqrt(1j*w * param.c1)
            param.bet = 1j*w * param.wo2 / ((1j*w + param.a) * 1j*w + param.b) + param.alp
            param.gam = np.sqrt((param.alp + 1j*w) / (param.bet + 1j*w))
            param.sig = param.gam * (param.bet + 1j*w)

        af_size = af.shape
        if len(af_size) == 1:
            nframe = 1
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
        else:
            nframe = af_size[1]
        transFun = np.zeros((lw, nframe)) + 1j*np.zeros((lw, nframe))
        Hf_cstr = np.zeros((lw, nframe)) + 1j*np.zeros((lw, nframe))

        if isinstance(loc, int) and nframe > 1:
            loc = loc * np.ones(nframe)
        if isinstance(loc, int):
            loc = [loc]
            
        if formants:
            fmt = np.zeros((param.nform, nframe))
        Grad = param.rho * w**2 / 2. / np.pi / param.c_s    
        
        if verbosity:
            disable = False
        else:
            disable = True            
        
        for kf in tqdm(range(nframe), disable=disable,
                       position=position, desc="Transfer function", leave=True):
            Avt = af[:, kf]
            lvt = lf[:, kf]
            Jrad = 8 * param.rho * w / 3. / np.pi**(1.5) * Avt[-1]**(-0.5)
            Zrad = Grad + 1j*Jrad

            A, _, C, _, _ = ChainMatrix(Avt, lvt, freq, param, meth)

            Hf = 1 / (-C * Zrad + A)
            transFun[:, kf] = Hf
            Hf_cstr[:, kf] = Hf

            if loc[kf] > 0:
                param.lg = 0.03
                param.Ag0 = 0.4e-4
                Zg = 12 * param.mu * param.lg**3 / param.Ag0**3 + \
                0.875 * param.rho / 2 / param.Ag0**2 + \
                1j*freq * 2 * np.pi * param.rho * param.lg / param.Ag0
                k_tmp = loc[kf]
                aup = Avt[k_tmp::-1]
                adown = Avt[k_tmp+1:]
                lup = lvt[k_tmp::-1]
                ldown = lvt[k_tmp+1:]
                Aup, Bup, Cup, Dup, _ = ChainMatrix(aup, lup, freq, param, meth)
                Adown, Bdown, Cdown, Ddown, _ = ChainMatrix(adown, ldown, freq, param, meth)
                Tf_front = 1 / (Cdown * Zrad + Ddown)
                Z_front = (Adown * Zrad + Bdown) / (Cdown * Zrad + Ddown)
                Z_back = (Aup * Zg + Bup) / (Cup * Zg + Dup)
                Hf_cstr[:, kf] = Tf_front * Z_back / (Z_front + Z_back)
            if formants:
                fmt[:, kf] = formant_from_tf(transFun[:, kf].squeeze(), 
                                         freq, 
                                         nform=param.nform)
        self.transfer_function = np.squeeze(transFun)
        self.transfer_function_constriction = np.squeeze(Hf_cstr)
        self.freq = freq
        if formants:
            self.formants = fmt
        return transFun, freq

    def computeformants(self, param=None, verbosity=False, position=0):
        """ Computes the resonance frequencies of the VT_Waveguide object.
         Use the ESPRIT method on the impulse response to estimate the poles of
         the transfer function"""

        if param is None:
            param = struct()
        if self.transfer_function is None:
            self.computetransferfunction(self, param)
        tf_size = self.transfer_function.shape
        if len(tf_size) == 1:
            nfreq = tf_size[0]
            nframe = 1
            extend = True
            self.transfer_function = self.transfer_function.reshape(-1, 1)
        else:
            nfreq, nframe = tf_size
            extend = False

        Hf = self.transfer_function
        Hf[0, :] = Hf[1, :]
        x = symmetricifft(Hf)
        if extend:
            x = x.reshape(-1, 1)
        x = x.shape[0] * x
        himp = x[:int(np.ceil(0.5 * len(x))), :]
        himp -= np.mean(himp, axis=0, keepdims=True)
        df = self.freq[2] - self.freq[1]
        K = 2 * (param.nform + 2)
        fmt = np.empty((param.nform, nframe))
        fmt[:] = np.nan
        if verbosity:
            disable = False
        else:
            disable = True   
        for kf in tqdm(range(nframe), disable=disable, position=position,
                       desc="Formants", leave=True):
            try:
                fk = esprit(himp[:, kf], df * 2 * (nfreq - 1), K)
                if len(fk) >= param.nform:
                    fmt[:, kf] = fk[:param.nform]
            except:
                pass
        self.formants = fmt
        return fmt

    def exportaf2json(self, fileName):
        """export the area function in a .JSON file"""
    
        af = self.area_function.area
        lf = self.area_function.length
    
        if hasattr(self, 'velopharyngeal_port'):
            vpo = self.velopharyngeal_port
        else:
            vpo = 0
        if hasattr(self, 'velum_area'):
            vArea = self.velum_area
        else:
            vArea = np.zeros_like(af)
    
        if len(af.shape) == 1:
            nFrame = 0
            nTubes = len(af)
        else:
            nTubes, nFrame = af.shape
    
        idx = fileName.find('.json')
        if idx >= 0:
            fileName = fileName[:idx]
    
        if isinstance(vpo, (float, int)) and nFrame > 0:
            vpo = vpo * np.ones(nFrame)
    
        if isinstance(vArea, (float, int)) and nFrame > 0:
            vArea = vArea * np.ones((nTubes, nFrame))
    
        if isinstance(vArea, (float, int)) and nFrame > 0:
            vArea = npmt.repmat(vArea.reshape(-1, 1), 1, nFrame)
    
        if nFrame == 0:
            fTmp = fileName + '.json'
            nFrame = 1
            vpo = [vpo]
            vArea = vArea * np.ones((nTubes, 1))
            af = af.reshape(-1, 1)
            lf = lf.reshape(-1, 1)
        for kF in range(nFrame):
            if nFrame > 1:
                fTmp = fileName + '_' + '%.5d' %(kF+1) + '.json'
            else:
                fTmp = fileName + '.json'
    
            idjson = {}
            idjson['VelumOpeningInCm'] = np.sqrt(vpo[kF]) * 1e2
            idjson['frameId'] = fTmp
            idjson['tubes'] = []
            for kT in range(nTubes):
                idjson['tubes'].append({
                    'area' : af[kT, kF] * 1e4,
                    'velumArea' : vArea[kT, kF] * 1e4,
                    'x' : lf[kT, kF] * 1e2
                })
        with open(fTmp, 'w') as f:
            json.dump(idjson, f, indent=4)
    
      
    def check_instance(self, class_type):
        """ check instance """
        return eval('isinstance(self, ' + class_type + ')')
    
def formant_from_tf(transfer_function, freq, nform=4):
    
    K = 2 * (nform + 2)
    if np.isinf(transfer_function[0]) or np.isnan(transfer_function[0]):
        transfer_function[0] = transfer_function[1]
    df = freq[2] - freq[1]
    x = symmetricifft(transfer_function)
    x = len(x) * x
    himp = x[:int(np.ceil(0.5 * len(x)))]
    himp -= np.mean(himp)
    try:
        fk = esprit(himp, df * 2 * (len(transfer_function) - 1), K)
        if len(fk) >= nform:            
            return fk[:nform]
        else:
            return np.array([np.nan]*nform)
    except:
        return np.array([np.nan]*nform)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:04:55 2022

@author: benjamin
"""

from .modeltools import check_model, normalize, crop

def articulatory_to_task(articulatory_parameters, model=None):
    linear_mapping_matrix, offset, factor = check_model(model)    
    p = normalize(articulatory_parameters, -3, 6, 0)
    t = (linear_mapping_matrix @ p.T).T
    t = crop(t, lower_bound=0, upper_bound = 1)
    
    t = normalize((linear_mapping_matrix @ p.T).T, offset, factor, direction=1)
    return crop(t, lower_bound=0)

    


        
    
    
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 26 11:16:32 2022

@author: benjamin
"""

import numpy as np
from shapely.geometry import LineString

class Contour:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=pointless-string-statement

    contours = None
    lips = None
    length = None
    tract_variables = {"tongue body" : {"degree" : None, "location" : None},
                       "tongue tip" : {"degree" : None, "location" : None},
                       "lip" : {"aperture" : None, "protrusion" : None},
                       "length": None}

    # Constructor method
    def __init__(self, *args):
        nargin = len(args)
        for k in range(0, nargin, 2):
            key, value = args[k:k+2]
            setattr(self, key, value)
            
    def compute_lip_variables(self):
        
        self.tract_variables["lip"]["aperture"] = get_lip_aperture(self)
        self.tract_variables["lip"]["protrusion"] = get_lip_protrusion(self)
            
    def compute_tongue_variables(self, origin=(10,10)):
        
        distance, thetas = self.get_distance(origin)
        for key_1, tr in zip(["tongue tip", "tongue body"],
                           [(110, 140), (0, 105)]):
            tr_var = extract_degloc(distance, thetas, tr)
            for key_2, value in zip(["degree", "location"], tr_var): 
                self.tract_variables[key_1][key_2] = value 
                
    def compute_tract_variables(self, c, origin=(10,10)):
        
        self.compute_lip_variables()
        self.compute_tongue_variables(origin)
        self.tract_variables["length"] = get_length(self, c)
                
    def get_distance(self, origin):
        
        thetas, xvecs = make_vecs(origin)
        lines = self.make_lines()
        distance_vec = []
        
        for n, theta in enumerate(thetas):
            second_line = make_angle_line(theta, xvecs, origin)
            intersections = [second_line.intersection(l) for l in lines]
            if all([ints.geom_type == "Point" for ints in intersections]):
                inner_point, outer_point = [get_points(ints) for ints in intersections]                       
                distance_vec.append(np.sqrt(np.sum((inner_point-outer_point)**2)))

        return distance_vec, thetas
    
    def interpolation(self, factor=1000):
        
        inner_x, inner_y, outer_x, outer_y = self.contours
        new_inner = interpol_contour(inner_x, inner_y, factor)
        new_outer = interpol_contour(outer_x, outer_y, factor)
        self.contours = (new_inner[0], new_inner[1], new_outer[0], new_outer[1])
        
    def make_lines(self):
        
        inner_x, inner_y, outer_x, outer_y = self.contours
        return (LineString(np.column_stack((inner_x, inner_y))),
                LineString(np.column_stack((outer_x, outer_y))))
    
def extract_degloc(distance, thetas, theta_range):
    
    th_deg = thetas * 180 / np.pi
    idx = [x for x in range(len(thetas)) if th_deg[x] >= theta_range[0] and 
              th_deg[x] <= theta_range[1]]

    distance_tract = [distance[i] for i in idx]
    constriction_degree = np.min(distance_tract)
    constriction_location = (180 - th_deg[idx[np.argmin(distance_tract)]])
    
    return constriction_degree, constriction_location

def extract_task_variables(contour, c, origin=(10,10)):
    
    try:
        contour.compute_tract_variables(c, origin)
        tracts = contour.tract_variables
        
        return [tracts["lip"]["aperture"], 
                  tracts["lip"]["protrusion"],
                  tracts["tongue tip"]["degree"], 
                  tracts["tongue tip"]["location"],
                  tracts["tongue body"]["degree"], 
                  tracts["tongue body"]["location"],
                  tracts["length"]]
    except:
        return [np.nan]*7

def get_length(contour, c):
    
    inner_x, inner_y, outer_x, outer_y = contour.contours
    x1 = -(np.diff(inner_x)+np.diff(outer_x))/100
    y1 = -(np.diff(inner_y)+np.diff(outer_y))/100
    d = 0.5*np.sqrt(x1**2 + y1**2)
    return np.sum(c*d)
    

def get_lip_aperture(contour):
    inner_x, inner_y, outer_x, outer_y = contour.contours
    y_upper_lip = outer_y[np.argmin(outer_x)]
    y_lower_lip = inner_y[np.argmin(inner_x)]
    return np.abs(y_upper_lip - y_lower_lip)

def get_lip_protrusion(contour):
    inner_x, inner_y, outer_x, outer_y = contour.contours
    x = np.sort(inner_x)
    return x[1] - x[0]

def get_points(intersection):
    return np.array([intersection.xy[k][0] for k in [0,1]])

def interpol_contour(x_coor, y_coor, D=1000):
    new_x = []
    new_y = []

    for n in range(len(x_coor)-1):
        x1, x2 = x_coor[n:n+2]
        y1, y2 = y_coor[n:n+2]
        
        x_vec = np.linspace(x1, x2, D)
        y_vec = np.linspace(y1, y2, D)
        if n == 0:
            s = 0
        else:
            s = 1
        
        for x, y in zip(x_vec[s:], y_vec[s:]):
            new_x.append(x)
            new_y.append(y)
            
    return new_x, new_y

def make_angle_line(theta, xvecs, origin):    
    a = np.sign(np.cos(theta)) * np.sqrt(1/np.cos(theta)**2 - 1)
    b = origin[1] - a * origin[0]
    xvec = xvecs[1-(a>=0)]            
    line = a * xvec + b
    return LineString(np.column_stack((xvec, line)))

def make_vecs(origin):
    return (np.linspace(0, 140, 1401) * np.pi/180, 
            (np.linspace(origin[0], 20, 5), 
            np.linspace(0, origin[0], 5)))


    
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 29 09:58:38 2022

@author: benjamin
"""

import json
import os
import numpy as np
from maedeep._contour import Contour
from maedeep._config import get_maedeep_model

def amo(px, py, qx, qy):
    return np.sqrt((px - qx)**2 + (py - qy)**2)
  
def check_model(model):
    
    if model is None:
        model = get_maedeep_model()
        
    if isinstance(model, str):
        if os.path.isfile(model):
            with open(model, "r") as file_id:
                return json.load(file_id)
        else:
            raise ValueError(model + " is not a valid file")
    elif isinstance(model, dict):
        return model
    else:
        raise ValueError("The model is not a valid object")
        
def contour_to_area_function(contour, c, iniva_tng, alph, beta):
    
    low_x, low_y, up_x, up_y = contour.contours
    lip_h, lip_w = contour.lips
    ivt = [{"x": x, "y": y} for x, y in zip(low_x, low_y)]
    evt = [{"x": x, "y": y} for x, y in zip(up_x, up_y)] 

    nb_p = len(ivt)
    af = [[]]*(nb_p)

    pv, qv, rv, sv, tv = cross_distance(ivt, evt)
    for i in range(1, nb_p-1):
        p = pv[i-1]
        q = qv[i-1]
        s = sv[i-1]
        r = rv[i-1]
        t = tv[i-1]
        a1 = 0.5*(p+s+t)
        a2 = 0.5*(q+r+t)
        s1 = np.sqrt(a1*(a1-p)*(a1-s)*(a1-t))
        s2 = np.sqrt(a2*(a2-q)*(a2-r)*(a2-t))
        x1 = ivt[i-1]["x"] + evt[i-1]["x"] - ivt[i]["x"] - evt[i]["x"]
        y1 = ivt[i-1]["y"] + evt[i-1]["y"] - ivt[i]["y"] - evt[i]["y"]
        d = 0.5*np.sqrt(x1**2 + y1**2)
        w = c*(s1 + s2)/d
        j = i + iniva_tng - 3
        af[i-1] = {"x": c*d,
                   "A": 1.4*alph[j]*w**beta[j]}
        
    af[nb_p-2:] = [{"A": np.pi * lip_h * lip_w * c**2,
                    "x": 0.5*(ivt[nb_p-2]["x"] - ivt[nb_p-1]["x"])*c} for i in range(2)] 

    af = [{"A": max([0.0001, x["A"]]),
           "x": max(0.01, x["x"])} for x in af]
    return {"area": [x["A"]*1e-4 for x in af], 
             "length": [x["x"]*1e-2 for x in af]}        
        
def cross_distance(ivt, evt):
    low_x, low_y, up_x, up_y = (np.array([x["x"] for x in ivt]),
                                np.array([x["y"] for x in ivt]),
                                np.array([x["x"] for x in evt]),
                                np.array([x["y"] for x in evt]))
    
    p = amo(low_x[1:], low_y[1:], low_x[:-1], low_y[:-1])
    q = amo(up_x[1:], up_y[1:], up_x[:-1], up_y[:-1])
    r = amo(low_x[:-1], low_y[:-1], up_x[:-1], up_y[:-1])
    s = amo(up_x[1:], up_y[1:], low_x[1:], low_y[1:])
    t = amo(up_x[1:], up_y[1:], low_x[:-1], low_y[:-1])
    return p, q, r, s, t

def vector_to_contours(articulatory_parameters, model, output_file=None):
    
    data = check_model(model)
        
    if isinstance(articulatory_parameters, list):
        articulatory_parameters = np.array(articulatory_parameters)
    if len(articulatory_parameters.shape) == 1:
        articulatory_parameters = articulatory_parameters.reshape(-1, 1)
    
    LIP, JAW, TNG, LRX = [int(data["semi-polar coordinates"]["number of parameters"][key])
                          for key in ["lip", "jaw", "tongue", "larynx"]]


    nvrs_lip, nvrs_tng, nvrs_lrx = [int(data[key]["parameters"]["number of variables"])
                                    for key in ["lips", "tongue", "larynx"]]

    A_lip, A_tng, A_lrx = [np.array(data[key]["statistics"]["factor"])
                           for key in ["lips", "tongue", "larynx"]]
    s_lip, s_tng, s_lrx = [(data[key]["statistics"]["standard deviation"])
                           for key in ["lips", "tongue", "larynx"]]
    u_lip, u_tng, u_lrx, u_wal = [(data[key]["statistics"]["mean"])
                           for key in ["lips", "tongue", "larynx", "wall"]]

    iniva_tng, lstva_tng = [int(data["tongue"]["parameters"][key])
                           for key in ["initial point", "last point"]]
    vtos = data["semi-polar coordinates"]["vector to semi-polar map"]
    igd, egd = [data["semi-polar coordinates"]["grids"][key] for key in ["interior", "exterior"]]
    ix0, iy0 = [data["semi-polar coordinates"]["origin"][key] for key in ["x", "y"]]
    inci_x, inci_y = [data["lips"]["statistics"]["upper lip"][key] for key in ["x", "y"]]
    inci_lip = data["semi-polar coordinates"]["incisor-lip distance"]
    vp_map = data["semi-polar coordinates"]["map coeff"]
    inci_lip_vp = inci_lip / vp_map
    DWIDTH, DHEIGHT = [data["semi-polar coordinates"]["display"][key]
                       for key in ["width", "height"]]
    TEKvt, TEKlip = [data["semi-polar coordinates"]["scale factor"][key]
                       for key in ["vocal tract", "lips"]]
    
    nb_params, nb_frame = articulatory_parameters.shape    
    contours = []
    
    for n in range(nb_frame):
        pa = articulatory_parameters[:, n].tolist()
        # pa = np.zeros(7)
        p = np.zeros(7)
        p[0] = pa[0]
        
        p[range(1, TNG+1)] = [pa[i] for i in range(1, TNG+1)]
        v_tng = np.zeros(nvrs_tng)
        v_lip = np.zeros(nvrs_lip)
        v_lrx = np.zeros(nvrs_lrx)
        for i in range(nvrs_tng):
            v = 0
            for j in range(JAW+TNG):
                v += A_tng[i, j]*p[j]
                v_tng[i] = s_tng[i]*v + u_tng[i]
                
        p[range(1, LIP+1)] = [pa[i+TNG] for i in range(1, LIP+1)]
        for i in range(nvrs_lip):
            v = 0
            for j in range(JAW+LIP):
                v += A_lip[i, j]*p[j]
                v_lip[i] = s_lip[i]*v + u_lip[i]
                if v_lip[i] < 0:
                    v_lip[i] = 0
                    
        p[range(1, LRX+1)] = [pa[i+TNG+LIP] for i in range(1, LRX+1)]
        for i in range(nvrs_lrx):
            v = 0
            for j in range(JAW+LRX):
                v += A_lrx[i, j]*p[j]
                v_lrx[i] = s_lrx[i]*v + u_lrx[i]
        
        ivt = [{"x": v_lrx[JAW] + ix0, 
                "y": v_lrx[JAW+1] + iy0}]
        evt = [{"x": v_lrx[JAW+2] + ix0, 
                "y": v_lrx[JAW+3] + iy0}]
        
        for i in range(iniva_tng, lstva_tng+1):
            j = i - iniva_tng
            v = min([v_tng[j+JAW], u_wal[j]])
            x1 = vtos[i]["x"]*v + igd[i]["x"]
            y1 = vtos[i]["y"]*v + igd[i]["y"]
            x2 = vtos[i]["x"]*u_wal[j] + igd[i]["x"]
            y2 = vtos[i]["y"]*u_wal[j] + igd[i]["y"]
            if i == iniva_tng:
                ivt.append({"x": (ivt[0]["x"]+x1)/2,
                            "y": (ivt[0]["y"]+y1)/2})
                evt.append({"x": (evt[0]["x"]+x2)/2,
                            "y": (evt[0]["y"]+y2)/2})
            ivt.append({"x": x1,
                        "y": y1})
            evt.append({"x": x2,
                        "y": y2})
        
        evt.append({"x": inci_x + ix0,
                    "y": inci_y + inci_lip_vp + iy0})
        ivt.append({"x": evt[-1]["x"],
                    "y": evt[-1]["y"] - v_lip[2]})
        
        evt.append({"x": evt[-1]["x"] - v_lip[1],
                    "y": evt[-1]["y"] })
        ivt.append({"x": evt[-1]["x"],
                    "y": ivt[-1]["y"]})
         
        P1 = np.array([[x["x"] for x in ivt],
                       [x["y"] for x in ivt],
                       [x["x"] for x in evt],
                       [x["y"] for x in evt]])
        
        lip_h = v_lip[2]/2
        lip_w = v_lip[3]/2
        
        contours.append(Contour("contours", P1, "lips", (lip_h, lip_w)))
        
    return contours


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:49:54 2022

@author: benjamin
"""

import numpy as np
import scipy

def in_hull_delaunay(p, hull):
    """
    Test if points in `p` are in `hull`

    `p` should be a `NxK` coordinates of `N` points in `K` dimensions
    `hull` is either a scipy.spatial.Delaunay object or the `MxK` array of the 
    coordinates of `M` points in `K`dimensions for which Delaunay triangulation
    will be computed
    """
    if not isinstance(hull, scipy.spatial.Delaunay):
        hull = scipy.spatial.Delaunay(hull)

    return hull.find_simplex(p)>=0

def var_range(x, axis=0):
    return np.max(x, axis=axis, keepdims=True) - np.min(x, axis=axis, keepdims=True)

def dnn_cost_function(generated, observed, relative=False):    
    if relative == True:
        return np.mean(np.abs((generated-observed)/observed))
    elif relative == False:
        return np.mean(np.abs((generated-observed)))
    elif relative == "range":
        observed_range = var_range(observed)
        return np.mean(np.abs((generated-observed)/observed_range))

def symmetricifft(x, nfft=None):
    """ returns the IFFT by symmetric inverse Fourier transform """
    if len(x.shape) == 1:
        x = x.reshape(-1,1)
    xconj = np.conj(x[-2:0:-1,:])
    xconc = np.concatenate((x, xconj), 0)
    if nfft is None:
        nfft = len(xconc)
    y = np.real(np.fft.ifft(xconc, nfft, axis=0))

    return np.squeeze(y)

def window(nwin, win_type):
    
    if win_type == 'hann':
        return scipy.signal.hann(nwin)
    elif win_type == 'hanning':
        return scipy.signal.hanning(nwin)
    elif win_type == 'hamming':
        return scipy.signal.hamming(nwin)
    else:
        raise ValueError('Wrong type of window')

def esprit(x, sr = 1, K = 12):
    """ estimate the frequency of the damped sinusoids that model x """
    
    idx = np.argmax(x)
    x = x[idx:]
    x = x - np.mean(x)
    
    M = int(min(100, np.floor(len(x) / 2)))
    Nl = len(x) - M + 1
    Nt = int(Nl / M)
    
    R = np.zeros((M,M))
    for k in range(Nt):
        deb = int(k * M)
        fin = int(deb + 2 * M - 1)
        xtmp = x[deb:fin]
    
        H = scipy.linalg.hankel(xtmp[0:M], xtmp[M - 1:])
        R += H.dot(H.T)
        
    u, s, d = np.linalg.svd(R)
    nx, ny = u.shape
    
    Up = u[1:,:K]
    Um = u[:-1,:K]
    Phi = np.linalg.pinv(Um).dot(Up)
    z,w = np.linalg.eig(Phi)
    freq = np.angle(z) / 2 / np.pi * sr
    
    return np.sort(freq[freq > 50])

def fill_nan(x):

    idxAll = np.arange(x.shape[0])
    idxVal = np.where(np.isfinite(x))
    try:
        f = scipy.interpolate.interp1d(idxAll[idxVal], x[idxVal], 
                                   kind='cubic', fill_value='extrapolate')
    except:
        f = scipy.interpolate.interp1d(idxAll[idxVal], x[idxVal], 
                                   kind='linear', fill_value='extrapolate')
    return np.where(np.isfinite(x), x, f(idxAll))

def fill_nan_and_zero(x):
    
    y = fill_nan(x)
    first_non_zero = np.argwhere(y>0)[0][0]
    last_non_zero = len(y) - np.argwhere(y[::-1]>0)[0][0] - 1
    y[:first_non_zero] = y[first_non_zero]
    y[last_non_zero:] = y[last_non_zero]
    
    return y

def nextpow2(x):
    return int(np.ceil(np.log2(np.abs(x))))
   
def ChainMatrix(af, lf, freq, param, meth='tmm',  Tf=None):
    lw = len(freq) 
    if Tf is None:
        Tf = np.tile(np.eye(2).reshape((1,2,2)),(lw, 1, 1))

    Zo = (af / (param.rho * param.c)).reshape(1, -1)

    if meth.lower() != 'cmp':
        om = param.freq * 2 * np.pi    
        S = 2 * np.sqrt(af * np.pi)        
        L = param.rho / af
        Celem = (Zo / param.c).squeeze()
        
    if meth.lower() == 'cmp':
        if param.loss:
            argh = param.sig.reshape(-1,1) @ lf.reshape(1, -1) / param.c
            gam = param.gam.reshape(-1, 1) @ np.ones((1, len(lf)))
        else:
            argh = (1j * om / param.c).reshape(-1,1) @ lf.reshape(1, -1)
            gam = 1        
        Amat = np.cosh(argh)
        Bmat = - gam * np.sinh(argh) * (np.ones((lw, 1)) @ (1 / Zo))
        Cmat = (- 1 / gam * np.sinh(argh)) * (np.ones((lw, 1)) @ Zo)
    
    for k in range(len(af)-1,-1,-1):        
        if meth.lower() == 'cmp':            
            A = Amat[:, k].squeeze()
            B = Bmat[:, k].squeeze()
            C = Cmat[:, k].squeeze()
        else:
            R = S[k]*np.sqrt(param.rho * param.mu * om) / (2 * np.sqrt(2) * af[k]**2)
            G = (param.adiabatic - 1) * S[k] / (param.rho * param.c**2) * np.sqrt(param.heat_cond * om / (2 * param.specific_heat * param.rho))
            if param.loss:
                invZw = 1 / (param.wr / S[k]**2 + 1j*om*param.wm / S[k]**2 + 1 / (1j * om * S[k]**2 / param.wc)) * param.wallyield
                gam = np.sqrt((R + 1j * om * L[k]) * (G + 1j * om * Celem[k] + invZw))
            else:
                gam = 1j * om / param.c

            A = np.cosh(gam * lf[k])
            B = -np.sinh(gam * lf[k]) / Zo.squeeze()[k]
            C = -Zo.squeeze()[k] * np.sinh(gam * lf[k]) 
        
        Tn = np.array([[A,B], [C,A]]).transpose((2,0,1))
        Tf = np.matmul( Tf, Tn )
        
    A = Tf[:, 0, 0].flatten()
    B = Tf[:, 0, 1].flatten()
    C = Tf[:, 1, 0].flatten()
    D = Tf[:, 1, 1].flatten()
    
    return A, B, C, D, Tf

def findelements(a, b):
    # find indices of elements in a that contains elements in b
    nB = len(b)
    idx = []
    
    for k in range(nB):
        bTmp = b[k]
        idxTmp = [i for i,x in enumerate(a) if x==bTmp]
        if idxTmp != []: 
            if idx == []:
                idx = idxTmp
            else:
                idx = np.concatenate((idx, idxTmp)).astype(int)
                
    return idx

def lininterp1(xi, yi, xo, axis=1):
    
    if type(xi) is list:
        f1 = scipy.interpolate.interp1d(xi, yi, 'linear', 
                                        fill_value='extrapolate')
    else:
        while axis >= len(yi.shape): 
            axis -= 1        
        f1 = scipy.interpolate.interp1d(xi, yi, 'linear',
                                        fill_value='extrapolate', axis=axis)
   
    return f1(xo)

def vtl_estimation(freq, c=343):
     
    sum_freq = 0
    for i, row in enumerate(freq.T):
        for F_ij in row:
            sum_freq += F_ij / (i + 0.5)
    return c * freq.size / (2*sum_freq)

def vtln(freq, target_length=0.16273747, c=343):
    
    if freq.shape[1] != 4 and freq.shape[0] == 4:
        freq = freq.T
    return (freq * vtl_estimation(freq) / target_length).T


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb  6 21:31:52 2019

@author: benjamin
"""

import os
import sys
import warnings
if not sys.warnoptions:
    warnings.simplefilter("ignore")
    os.environ["PYTHONWARNINGS"] = "ignore"
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' 

from ._set import *
from .signaltools import *
from .gmmtools import *
from ._areafunction import *
from ._contour import *
from ._waveguide import *
from ._config import *
# from .dnn import *
# from .parametric import *
